#!/usr/bin/perl -w
use strict;
#use diagnostics;

use lib "$ENV{CMISS_EXAMPLES}/common/cmiss_type1";
use test_cmiss;

my %options;
my $usage = "Usage: $0 [-h] [-e example_path] [-f example_list_filename]\n";
my $example_path = "";
my $example_list_filename;

#Keep all the flags other than -e so we can pass them along.
my @flags;
for( my $i = 0; $i < @ARGV; $i++ ) {
    my $arg = $ARGV[$i];
    if( $arg eq '-e' ) {
	$i++; # skip argument associated with -e
    } else {
	push @flags, $arg;
    }
}

#Use our own version of getopts that ignores unknown options so that
#these can be used by subsequent programs;
if (!(test_cmiss::getopts('he:f:', \%options)))
  {
    print $usage;
    exit;
  }

sub abort {
    die "$0: ",@_;
}
sub mywarn {
    warn "$0 warning: ",@_;
}

$SIG{__DIE__} = \&abort;
$SIG{__WARN__} = \&mywarn;

for (keys %options)
  {
    if (m%^h$%)
      {
	 print $usage;
	 exit;
      }
    elsif (m%^e$%)
      {
	#The relative path to this example i.e. a/a1
	$example_path = $options{$_};
      }
    elsif (m%^f$%)
      {
	#A filename to which is written a list of all parsed examples
	$example_list_filename = $options{$_};
      }
  }

#Where all the examples are stored on disk i.e. /product/cmiss/examples
my $example_location = $ENV{CMISS_EXAMPLES};
#Script/template path
my $template_path = "$example_location/common/cmiss_type1";
#Template file
my $template_file = "$template_path/template_node.html";
my $readdir_grep_string = q/^(\.|CVS)/;

#Get environment variables, use a sensible dummy if not found.
my $cmiss_www_example_url = "file:$example_location/";
if ($ENV{CMISS_WWW_EXAMPLE_URL})
  {
	 $cmiss_www_example_url = $ENV{CMISS_WWW_EXAMPLE_URL};
  }
my $cmiss_www_url = "/\$CMISS_WWW_URL";
if ($ENV{CMISS_WWW_URL})
  {
	 $cmiss_www_url = $ENV{CMISS_WWW_URL};
  }

my $tree_ref = run_html_tree($example_path);

sub list_tree {
    my ($tree_ref) = @_;

    my $name = $tree_ref->{name};
    my $example_path = $tree_ref->{example_path};

    my $link_text = (exists $tree_ref->{subtrees} ? "Node" : "Example") .
	(length $example_path ? " $example_path: " : ' ') . $name;

    print EXAMPLE_LIST
      test_cmiss::html_link
	  ( $cmiss_www_example_url, $example_location,
	    length $example_path ? "$example_path/index.html" : 'index.html',
	    $link_text );

    my $subtrees_ref = $tree_ref->{subtrees};
    if( $subtrees_ref && %$subtrees_ref ) {

	print EXAMPLE_LIST "<UL>\n";

	foreach my $subdir (sort {$a cmp $b} keys %$subtrees_ref) {
	    print EXAMPLE_LIST '<LI>';
	    list_tree($subtrees_ref->{$subdir});
	}

	print EXAMPLE_LIST "</UL>\n";
    }
}

if ($example_list_filename)
  {
    open (EXAMPLE_LIST, ">$example_list_filename");
    print EXAMPLE_LIST "<HTML>\n";
    print EXAMPLE_LIST "<H1>Example List</H1>\n";

    list_tree($tree_ref);

    my $date = localtime(time());

    print EXAMPLE_LIST "<p>Generated: $date\n";
    print EXAMPLE_LIST "</HTML>\n";
    close EXAMPLE_LIST;
  }

sub process_node {
    my ($name,$example_path,$subtrees_ref) = @_;

    my $example_dir = length $example_path ?
	"$example_location/$example_path" : $example_location;

    my $example_tree_list =
      do {
	my $list = '';
	my @subpath_tokens = split "/", $example_path;
	# Prepend the help page if we know where it is.
	# ??? Should we use the bioeng home page if we don't? 
	if ( exists $ENV{CMISS_WWW_URL} )
	  {
	    $list .= <<HTML;
<a href="$ENV{CMISS_WWW_URL}help/index.html">CMISS Help</a> /
HTML
	    unshift @subpath_tokens, 'Examples'
	  }
	else
	  {
	    unshift @subpath_tokens, 'CMISS Examples'
	  }
	my $example_name = pop @subpath_tokens;
	# References to parent directories
	while ( @subpath_tokens > 0 )
	  {
	    my $dots = '../' x @subpath_tokens;
	    my $token = shift @subpath_tokens;
	    $list .= <<HTML;
<a href="${dots}index.html">$token</a> /
HTML
	  }
	# ??? Do we still want to print the name if there are no links?
	$list .= $example_name;
	$list;
      };

    open FILE, ">$example_dir/index.html"
	or die "can't open $example_dir/index.html: $!\n";

    open (INPUT, "<$template_file") || die "can't open $template_file: $!\n";
    
    while(<INPUT>)
      {
# KAT: Removed because not used; use a relative link if needed.
# 	s%\$\{CMISS_WWW_EXAMPLE_URL\}%$cmiss_www_example_url%g;
# Not used:
# 	s/\$example_name/$example_name/g;
# 	s%\$\{CMISS_HELPURL\}%${cmiss_www_url}help/%g;
#         s%\$\{CMISS_WWW_URL\}%$cmiss_www_url%g;

	s/<EXAMPLE_TREE_LIST>/$example_tree_list/;

	#All the other tags and tokens work on a line by line basis.
	#Therefore an <if IMAGES> tag should be on a line by itself
	#as well as the corresponding </if>
    
	#Allow conditional sections in the template.
	#<if *> is only recognised when blocked against the left margin.
	#i.e. <if IMAGES>
	#     This is only displayed if the images directory exists.
	#     </if>
    if (m%^<if [^>]*>%)
      {
	my $iffalse = 0;
	if (m%^<if IMAGES>%)
	  {
	    if ( ! -d "$example_dir/web_data/images" )
	      {
		$iffalse = 1;
	      }
	  }
	if ($iffalse)
	  {
	    my $ifdepth = 1;
	    while ($ifdepth && ($_ = <INPUT>))
	      {
		if (m%^<if %)
		  {
		    $ifdepth++;
		  }
		elsif (m%^</if%)
		  {
		    $ifdepth--;
		  }
	      }
	  }
      }
    elsif (m%^</if[^>]*>%) {
	# Discard
    }
    
    #Substitute strings for content
    elsif (m/\$name_file_goes_here/) {
	print FILE "$name\n" if defined $name;
    }
    elsif (m/\$description_file_goes_here/)
      {
	if (-f "$example_dir/web_data/description.html")
	  {
	    open (DESCRIPTION_FILE, "<$example_dir/web_data/description.html")
	      || die "can't open $example_dir/web_data/description.html: $!\n";
	    while (<DESCRIPTION_FILE>) {
# KAT: Removed because not used; use a relative link if needed.
# 		s%\$\{CMISS_WWW_EXAMPLE_URL\}%$cmiss_www_example_url%g;
		s%\$\{CMISS_HELPURL\}%${cmiss_www_url}help/%g;
		s%\$\{CMISS_WWW_URL\}%$cmiss_www_url%g;
		print FILE;
	    }
	    close DESCRIPTION_FILE;
	}
    }
# The format for the files in the file list is expected following the tag.
    elsif (m/\$directory_list_goes_here/)
    {
	my $format = substr ($_, length ("\$directory_list_goes_here"));
	foreach my $subdir (sort {$a cmp $b} keys %$subtrees_ref) {
	    my $output = $format;
	    my $name = $subtrees_ref->{$subdir}->{name};

	    my $link = test_cmiss::html_link
		('',$example_dir,"$subdir/index.html",$name);

	    $output =~ s%\$directory%$subdir%g;
	    $output =~ s%\$link%$link%g;
	    print FILE $output;
	}
    }
elsif (m/\$generated_time_stamp/)
  {
    my $output = localtime(time());
    s/\$generated_time_stamp/$output/;
    print FILE $_;
  }
	elsif (m/\$input_modified_time_stamp/)
	{
	    # Look through the cmiss_input files and the web_data files
	    # for the most recent

	    # Get list of input files.
	    my @input_files;
	    if( -d (my $input_dir = "$example_dir/web_data") ) {
		opendir DIR, $input_dir
		    or die "can't open $input_dir:$!\n";
		# List all files except for hidden ones.
		push @input_files, map {"$input_dir/$_"}
		sort {$a cmp $b} grep {!/$readdir_grep_string/} readdir DIR;
		closedir DIR;
	    }

	    my $recent = 0;
	    foreach my $file (@input_files)
	    {
		my $mtime = (stat "$file")[9];
		$recent = $mtime if $mtime > $recent;
	    }

	    my $output = localtime $recent;
	    s/\$input_modified_time_stamp/$output/;
	    print FILE $_;
	}
	else
	  {
	    print FILE $_;
	  }
      }

    close INPUT;
    close FILE;
  }

#Recurse down tree.
sub run_html_tree
  {
    my $example_path = shift;
    my $example_dir = $example_location . "/" . $example_path;

    # builds html files for the tree below example_path and returns a
    # reference to a hash of info from the tree.  Possible keys in the
    # hash are name, example_path, and subdirs.  If the subtrees key
    # exists, its value is a hash of subtree hashes indexed by subdir
    # name; if it doesn't exist this is a leaf (example) directory.
    # Returns undef if nothing interesting was found.

    my %tree;
    
    if (test_cmiss::is_leaf_directory("$example_dir")) {
	my $script = "$example_dir/control/html";
	if (-x $script) {
	    # Alter the umask for generated files according to read
	    # permissions on the script.  We should really remove group
	    # read privilege if other read is not set and we are in a
	    # different group but CVS doesn't bother with this so
	    # there doesn't seem much point.
	    my $mode = ( stat _ )[2];
	    my $old_umask = umask;
	    umask( $old_umask | ( ~$mode & 0044 ) );

	    # get example name
	    if (-f "$example_dir/web_data/name.txt") {
		open (NAME_FILE, "<$example_dir/web_data/name.txt")
		  || die "can't open $example_dir/web_data/name.txt: $!\n";
		my $name = <NAME_FILE>; chomp $name;
		close NAME_FILE;
		$tree{name} = $name;

	    }
	    else {
		warn "no name.txt found for $example_path\n";
		$tree{name} = '*nameless*';
	    }

	    # produce index.html file
	    my @command = ($script,'-e',$example_path,@flags);
	    system @command;
	    $? != -1 or die "can't run $script: $!\n";
	    $? & 0xFF and die "signal $? in @command\n";
	    $? == 0 or warn "unable to successfully complete @command: $?\n";

	    # return umask to original
	    umask $old_umask;
	  }
    }
    else {

	# get node name
	my $name;
	if( -f "$example_dir/web_data/node_name.txt" )
	  {
	    open (NAME_FILE, "<$example_dir/web_data/node_name.txt")
	      || die "can't open $example_dir/web_data/node_name.txt: $!\n";
	    $name = <NAME_FILE>; chomp $name;
	    close NAME_FILE;
	  }
	
	#Look further down.
	opendir(DIR, "$example_dir") || die "can't opendir $example_dir: $!\n";
	my @subdirectories =
	    sort { $a cmp $b } grep {
		! /$readdir_grep_string/ && -d "$example_dir/$_"
		} readdir(DIR);
	closedir DIR;

	my %subtrees;
	for my $subdir (@subdirectories) {
	    if( $subdir =~ m%^web_data$% ) {
		#Ignore
	    }
	    else {
		my $subtree_ref =
		    run_html_tree( ( length $example_path > 0 ?
				     "$example_path/$subdir" : $subdir ) );
		# include this subtree if something interesting was found.
		$subtrees{$subdir} = $subtree_ref if defined $subtree_ref;
	    }
	}

	# include a subtrees key in %tree and index.html if subtrees
	# or node_name was found.

	if( %subtrees || $name ) {

	    $tree{subtrees} = \%subtrees;
	    $tree{name} = $name || '*unnamed*';

	    # make index.html for this node
	    process_node ($name, $example_path, \%subtrees );
	}
    }

    if( %tree ) {
	$tree{example_path} = $example_path;
	return \%tree;
    } else {
	# nothing interesting to report
	return;
    }
  }

### Local Variables: 
### tab-width: 8
### End: 
