eval 'exec perl -S $0 ${1+"$@"}'
    if 0;

use strict;
#use diagnostics;

$0 =~ s%^.*/%%; # remove path from warning messages
BEGIN {
    $^W = 1; # turn on warnings
    $SIG{__WARN__} = sub { warn "$0 warning: ",@_ };
}

use FileHandle;

sub dirname ($) {
    # return the directory part of a filename
    $_[0] =~ m'^(.*)/[^/]*$' ? $1 : '.';
}
# use modules in the same directory as this file if they exist.
use lib dirname __FILE__, "$ENV{CMISS_EXAMPLES}/common/cmiss_type1";
use test_cmiss;

use Getopt::Std;

my %options;
my $usage = "Usage: $0 [-h] [-F] [-e example_path] [-o output_directory] [-p programmer_email_regexp] [-n notify_all_email] [-f notify_all_filename]\n";
my $seconds_in_category = 86400; # 1 day
my $oldest_category = 365; # days
my $example_path = "";
my $programmer_name_regexp;
my @notify_all_list;
my $notify_filename;
my $include_ftnchek;
my $output_directory;

if (!(getopts('he:p:n:f:o:F', \%options)))
  {
    die $usage;
  }

for (keys %options)
  {
    if (m%^h$%)
      {
	 print $usage;
	 exit;
      }
    elsif (m%^e$%)
      {
	#The relative path to this example e.g. a/a1
	$example_path = $options{$_};
      }
    elsif (m%^p$%)
      {
	#A regular expression for the programmers to whom mail is sent 
	$programmer_name_regexp = $options{$_};
      }
    elsif (m%^n$%)
      {
	#This space separated list of email addresses receives email for all
	#examples this mail compiles for ignoring the notify list.
	@notify_all_list = split ' ', $options{$_};
      }
    elsif (m%^f$%)
      {
	#This file will be created and results for every example written into.
	$notify_filename = $options{$_};
      }
    elsif (m%^o$%)
      {
	#Want to use a specified output directory - probably different to the default
	#of having the output in the example tree
	$output_directory = $options{$_};
      }
    elsif (m%^F$%)
      {
	#Include ftnchek results for notify_all_list and notify_filename.
	$include_ftnchek = 1;
      }
  }

local *Stdout;
#If no output is requested then send output to STDOUT.
unless( defined $programmer_name_regexp
	or @notify_all_list
	or defined $notify_filename ) {
    # save STDOUT for mail file and redirect progress messages to STDERR
    open Stdout, ">&STDOUT" or die "can't dup STDOUT: $!\n";
    open STDOUT, ">&STDERR" or die "can't dup STDERR: $!\n";
    $notify_filename = '&Stdout';
}

#Where all the examples are stored on disk e.g. /product/cmiss/examples
my $example_location = $ENV{CMISS_EXAMPLES};
#If no output directory has been specified, default to the standard
#example location path
if (! defined $output_directory)
{
    $output_directory = $example_location;
}

my $ftnchek_location = $ENV{CM_PARSE_FTNCHEK} || "$ENV{CMISS_ROOT}/cmiss_utils/www/fortran_check";
#Get environment variables, use a sensible dummy if not found.
my $cmiss_www_example_url =
  $ENV{CMISS_WWW_EXAMPLE_URL} || "file:$example_location/";
my $cmiss_www_help_url = $ENV{CMISS_WWW_HELP_URL} || "file:$ENV{CMISS_ROOT}/www/help/";
my $cmiss_www_ftnchek_url = "${cmiss_www_help_url}errors/";

#Script/template path
my $template_path = "$example_location/common/cmiss_pre2000";
my $readdir_grep_string = q/^(\.|CVS)/;

#=============================================================================

# Set up a protect mechanism that provide a finish that always
# executes and a catch for dies.

sub protect ( &% ) {
    my ($try,%ops) = @_;
    eval {
 	local $SIG{__DIE__}; # default die
	$try->();
    };
    exists $ops{finish} and do {
	# protect the finish also while retaining @_;
	my $save_at = $@;
	eval { $ops{finish}->() };
	$@ = $save_at . $@;
    };
    $@ and exists $ops{catch} ? $ops{catch}->() : die $@;
}
sub finish (&) { finish => $_[0] };
sub catch (&) { catch => $_[0] };

#=============================================================================

if ($notify_filename)
{
    #Append the filename as a notify all user.
    push @notify_all_list,$notify_filename;
}

# get a complete list of failures
my ($notify_hash_ref,$failure_list_ref) =
  test_cmiss::make_mail_trees( $output_directory, $example_location,
			       [$example_path], $programmer_name_regexp,
			       \@notify_all_list);

my $end_parse_time = time;
my $date = localtime $end_parse_time;

my (@cm_notify_list,@cmgui_notify_list);
if( defined $programmer_name_regexp )
  {
    test_cmiss::read_notify_list("$example_location/common/cmiss_type1/allcm.notify_list", 
				 \@cm_notify_list);
    test_cmiss::read_notify_list("$example_location/common/cmiss_type1/allcmgui.notify_list", 
				 \@cmgui_notify_list);
  }

# Get fortran check errors if necessary.
# if( my @ftnchek_notify_list = @notify_all_list ) # !!! for testing !!!
if( my @ftnchek_notify_list =
      ( defined $programmer_name_regexp ?
	  grep { m/$programmer_name_regexp/ } @cm_notify_list
	    : (),
	$include_ftnchek ?
	  @notify_all_list
	    : () ) )
  {
    # Don't die if this fails.
    protect
      {
	local *DIR;
	opendir DIR, $ftnchek_location or
	  die "can't opendir $ftnchek_location: $!\n";

	my @subdirs =
	  sort { $a cmp $b }
	    grep
	      {
		! m/$readdir_grep_string/
		  and -f "$ftnchek_location/$_/summary.txt";
	      } readdir DIR;
	closedir DIR;

	foreach my $subdir (@subdirs)
	  {
	    my $summary_file = "$ftnchek_location/$subdir/summary.txt";

	    local *SUMMARY;
	    open SUMMARY, $summary_file or die "can't open $summary_file: $!\n";

	    while( defined( my $line = <SUMMARY> ) )
	      {
		chomp $line;
		my ($time_stamp,$filename,$title) = split ' ', $line, 3;

		# Skip old errors for now.
		$time_stamp eq '2003-05-01,03:29' and next;

		push @$failure_list_ref,
		  {
		    'ftnchek file' => "$subdir/$filename",
		    'name' => $title,
		    'break time' => test_cmiss::interpret_time_stamp $time_stamp,
		    'notify list' => \@ftnchek_notify_list
		   };
	      }
	    close SUMMARY;
	  }
      }
    catch
      {
	warn "not parsing ftnchek errors: $@"
      };
  }

# Store in the notify_hash (keyed by mail person) the failed examples
# categorized by date.

foreach my $failure_ref (@$failure_list_ref) {
    # Create a date category by dividing time into days since the
    # example tree parsing finished.
    my $date_category = do {
	use integer;
	($end_parse_time - $failure_ref->{'break time'})
	    / $seconds_in_category;
    };
    if( $date_category > $oldest_category ) {
	$date_category = $oldest_category;
    }
    # record failure against person and date category
    foreach my $notify_name ( @{$failure_ref->{'notify list'}} ) {
	push @{$notify_hash_ref->{$notify_name}->{$date_category}->{failures}},
	    $failure_ref;
    }
    # delete failure records no longer required
    delete $failure_ref->{'break time','notify list'};
}

print "Mail list is: @{[ keys %$notify_hash_ref ]}\n";
while( my ($mail_person,$category_hash_ref) = each %$notify_hash_ref ) {

    my $to_notify_filename =
	$notify_filename && ($mail_person eq $notify_filename);
    my $example_failure = defined $category_hash_ref;

    # determine the min and max time in each category
    if( $example_failure ) {
	while( my ($date_category,$info_hash_ref) = each %$category_hash_ref )
	{
	    my @times = map {
		$_->{'break time'}
	    } @{$info_hash_ref->{failures}};
	    @{$info_hash_ref}{'min','max'} = (&min(@times),&max(@times));
	}
    }
    # sorted list of date categories
    my @sorted_date_categories = sort { $a <=> $b} keys %$category_hash_ref;

    if( $to_notify_filename )
      {
	open (MAIL, ">$notify_filename");
      }
    else
      {
	open (MAIL, "|mail $mail_person");
      }

    #Determine membership
    #Reset all flags
    my $cmgui_programmer = 0;
    my $cm_programmer = 0;
    if (grep {m"$mail_person"} @cmgui_notify_list)
      {
	$cmgui_programmer = 1;
      }
    if (grep {m"$mail_person"} @cm_notify_list)
      {
	$cm_programmer = 1;
      }

    #Determine failures
    #Reset all flags
    my $cmgui_compile_failure = 0;
    my $unemap_compile_failure = 0;
    if ($cmgui_programmer)
      {
		  if (! -z "$ENV{CMISS_ROOT}/cmgui/mailfiles/programmer.mail")
			 {
				$cmgui_compile_failure = 1;
			 }
		  if (! -z "$ENV{CMISS_ROOT}/cmgui/mailfiles/unemap_programmer.mail")
			 {
				$unemap_compile_failure = 1;
			 }
      }

    #Header
    my $subject;
	 my @failures = ();
	 my $number_of_failures;
    if ($cmgui_compile_failure)
      {
		  push(@failures, "Cmgui compilation");
		}
    if ($unemap_compile_failure)
      {
		  push(@failures, "Unemap compilation");
		}
    if ($example_failure)
      {
		  push(@failures, "Cmiss $example_path tests");
		}
	 $number_of_failures = scalar @failures;
	 if ($number_of_failures)
		{
		  my $i;

		  for($i = 0 ; $i < $number_of_failures - 2 ; $i++)
			 {
				#List items
				$failures[$i] .= ", ";
			 }
		  if ($i < $number_of_failures - 1)
			 {
				#Second to last
				$failures[$i] .= " and ";
				$i++;
			 }
		  $subject = 'Problems in ' . join ('', @failures);
		}
	 else
		{
		  if ($cmgui_programmer)
			 {
				$subject = "Cmgui and Unemap compilations and Cmiss $example_path tests successful";
			 }
		  else
			 {
				$subject = "Cmiss $example_path tests successful";
			 }
		}

#     print "\n$mail_person: $subject\n";
    unless( $to_notify_filename ) {
	print MAIL "To: $mail_person\n";
	print MAIL "Subject: $subject\n";
	print MAIL "MIME-Version: 1.0\n";
	print MAIL "Content-type: text/html\n";
    }
    print MAIL <<HTML;
<HTML>
<HEAD>
<TITLE>$subject - $date</TITLE>
</HEAD>
HTML
    # cmgui compilation
    if( $to_notify_filename || $cmgui_programmer )
      {
	if ($cmgui_compile_failure)
	  {
	    if (-f "$ENV{CMISS_ROOT}/cmgui/mailfiles/programmer.mail")
	      {
		print MAIL "<H3>Cmgui compilation failure:</H3>";
		print MAIL "<pre>\n";
		open (CMGUI_COMPILATION_FILE, "<$ENV{CMISS_ROOT}/cmgui/mailfiles/programmer.mail")
		  || die "can't open $ENV{CMISS_ROOT}/cmgui/mailfiles/programmer.mail: $!\n";
		while(<CMGUI_COMPILATION_FILE>)
		  {
		    print MAIL $_ ;
		  }
		close CMGUI_COMPILATION_FILE;
		print MAIL "</pre>\n";
		print MAIL "<HR>\n";
	      }
	  }
	if ($unemap_compile_failure)
	  {
	    if (-f "$ENV{CMISS_ROOT}/cmgui/mailfiles/unemap_programmer.mail")
	      {
		print MAIL "<H3>Unemap compilation failure:</H3>";
		print MAIL "<pre>\n";
		open (UNEMAP_COMPILATION_FILE, "<$ENV{CMISS_ROOT}/cmgui/mailfiles/unemap_programmer.mail")
		  || die "can't open $ENV{CMISS_ROOT}/cmgui/mailfiles/unemap_programmer.mail: $!\n";
		while(<UNEMAP_COMPILATION_FILE>)
		  {
		    print MAIL $_ ;
		  }
		close UNEMAP_COMPILATION_FILE;
		print MAIL "</pre>\n";
		print MAIL "<HR>\n";
	      }
	  }
    }

    # cm compilation
    if( $to_notify_filename || $cm_programmer )
    {
	if( -s "$ENV{CMISS_ROOT}/www/help/compilation/irix_cm_fail.log" ) {
	    print MAIL "<a href=\"${cmiss_www_help_url}compilation/irix_cm_fail.log\">IRIX cm compilation FAILED !</a>\n";
	    print MAIL "<HR>\n";
	}
	if( -s "$ENV{CMISS_ROOT}/www/help/compilation/i686-linux_cm_fail.log" ) {
	    print MAIL "<a href=\"${cmiss_www_help_url}compilation/i686-linux_cm_fail.log\">Linux cm compilation FAILED !</a>\n";
	    print MAIL "<HR>\n";
	}
	if( -s "$ENV{CMISS_ROOT}/www/help/compilation/rs6000-aix_cm_fail.log" ) {
	    print MAIL "<a href=\"${cmiss_www_help_url}compilation/rs6000-aix_cm_fail.log\">AIX cm compilation FAILED !</a>\n";
	    print MAIL "<HR>\n";
	}
    }

    # Recent CMISS examples
    if( $example_failure ) {
	my $first_category = $sorted_date_categories[0];
	if( $first_category eq '0' ) {
	    # examples broken recently or pending
	    &html_failures( $category_hash_ref->{0} );
	    # remove reported category
	    delete $category_hash_ref->{0};
	    shift @sorted_date_categories;
	} else {
	    my $since_date =
		localtime( $category_hash_ref->{$first_category}->{max} );
	    print MAIL
"<H3>No errors have been discovered since $since_date.</H3>\n";
   
	}
    } else {
	print MAIL "<H3>Successful completion of all $example_path examples you are notified about.</H3>\n";
    }
    print MAIL "<HR>\n";

    # Additional attachments for cmgui programmers
    if ($cmgui_programmer)
      {
	if (-f "$ENV{CMISS_ROOT}/cmgui/utilities/cmgui_cvs_changes.log")
	  {
	    if (-z "$ENV{CMISS_ROOT}/cmgui/utilities/cmgui_cvs_changes.log")
	      {
		print MAIL "<H3>No recent cmgui source changes</H3>";
		print MAIL "<HR>\n";
	      }
	    else
	      {
		print MAIL "<H3>Cmgui source changes:</H3>";
		open (CMGUI_CHANGES_FILE, "<$ENV{CMISS_ROOT}/cmgui/utilities/cmgui_cvs_changes.log")
		  || die "can't open $ENV{CMISS_ROOT}/cmgui/utilities/cmgui_cvs_changes.log: $!\n";
		while(<CMGUI_CHANGES_FILE>)
		  {
		    print MAIL $_ ;
		  }
		close CMGUI_CHANGES_FILE;
		print MAIL "<HR>\n";
	      }
	  }
      }

    # Additional attachments for cm programmers
    if ($cm_programmer)
      {
	my $changesfile = "$ENV{CMISS_ROOT}/www/help/programmer/cm_changes.html";
	my $html = 1;

	if (-f "$changesfile")
	  {
	    print MAIL "<H3>Cm source changes:</H3>";
	    $html or print MAIL "<pre>\n";
	    open (CHANGES_FILE, "<$changesfile")
		|| die "can't open $changesfile: $!\n";
	    while(<CHANGES_FILE>)
	      {
		print MAIL $_ ;
	      }
	    close CHANGES_FILE;
	    $html or print MAIL "</pre>\n";
	    print MAIL "<HR>\n";
	  }
      }

    if( ($cmgui_programmer || $cm_programmer)
	&& -f "$example_location/example_cvs_changes.log" )
      {
	if (-z "$example_location/example_cvs_changes.log")
	  {
	    print MAIL "<H3>No recent changes to examples in CVS repository</H3>";
	    print MAIL "<HR>\n";
	  }
	else
	  {
	    print MAIL "<H3>Example CVS repository changes:</H3>";
	    open (EXAMPLES_CHANGES_FILE, "<$example_location/example_cvs_changes.log")
	      || die "can't open $example_location/example_cvs_changes.log: $!\n";
	    while(<EXAMPLES_CHANGES_FILE>)
	      {
		print MAIL $_ ;
	      }
	    close EXAMPLES_CHANGES_FILE;
	    print MAIL "<HR>\n";
	  }
      }

    # Older CMISS examples
    if( $example_failure ) {
	# sort with most recent first
	foreach my $date_category ( @sorted_date_categories ) {
	    &html_failures( $category_hash_ref->{$date_category} );
	}
	print MAIL "<HR>\n";
    }

    print MAIL "<p>Mail generated: $date\n";
    print MAIL "</HTML>\n";
    
    close MAIL;
}

sub html_failures {
    my ($info_hash_ref) = @_;
    my ($failure_list_ref,$min_time,$max_time) =
	@{$info_hash_ref}{'failures','min','max'};

    if( $min_time == $max_time ) {
	my $date = localtime( $min_time );
	print MAIL "<H3>The following errors were discovered at $date:</H3>\n";
    } else {
	my $from_date = localtime( $min_time );
	my $to_date = localtime( $max_time );
	print MAIL "<H3>The following errors were discovered between $from_date and $to_date:</H3>\n";
    }
    print MAIL "<OL>\n";
    foreach my $failure_ref (@$failure_list_ref)
      {
	print MAIL '<LI>';
	if( exists $failure_ref->{'ftnchek file'} )
	  {
	    print MAIL
	      test_cmiss::html_link
		  ($cmiss_www_ftnchek_url,$ftnchek_location,
		   "$failure_ref->{'ftnchek file'}",
		   "$failure_ref->{name}");
	  }
	else
	  {
	    print MAIL
	      test_cmiss::html_link
		  ($cmiss_www_example_url,$example_location,
		   "$failure_ref->{'example path'}/index.html",
		   "$failure_ref->{'example path'}: $failure_ref->{name}");
	  }
	my $failed_versions_ref = $failure_ref->{'failed versions'} || [];
	my $incomplete_versions_ref = $failure_ref->{'incomplete versions'} || [];
	if( @$failed_versions_ref ) {
	    print MAIL "<br>Failed versions: @$failed_versions_ref\n";
	}
	if( @$incomplete_versions_ref ) {
	    print MAIL "<br>Incomplete versions: @$incomplete_versions_ref\n";
	}
    }
    print MAIL "</OL>\n";
}

sub min {
    my $min;
    foreach my $current (@_) {
	unless( defined $min and $current >= $min ) { $min = $current }
    }
    return $min;
}

sub max {
    my $max;
    foreach my $current (@_) {
	unless( defined $max and $current <= $max ) { $max = $current }
    }
    return $max;
}

### Local Variables: 
### mode: perl
### tab-width: 8
### End: 
