#!/usr/bin/perl -w
use strict;
#use diagnostics;

use lib "$ENV{CMISS_EXAMPLES}/common/cmiss_type1";
use test_cmiss;

use File::Path;

my %options;
my $usage = "Usage: $0 [-h] [-e example_path] [-t]\n";
my $example_path = "";
my $cm_testing = 0;

#Use our own version of getopts that ignores unknown options so that
#these can be used by subsequent programs;
if (!(test_cmiss::getopts('he:t', \%options)))
  {
    print $usage;
    exit;
  }

for (keys %options)
  {
    if (m%^h$%)
      {
	 print $usage;
	 exit;
      }
    elsif (m%^e$%)
      {
	#The relative path to this example i.e. a/a1
	$example_path = $options{$_};
      }
    elsif (m%^t$%)
      {
	#Add files for cm_testing to the example
	$cm_testing = $options{$_};
      }
  }

my $readdir_grep_string = q/^(\.|CVS)/;
#Where all the examples are stored on disk i.e. /product/cmiss/examples
my $example_location = $ENV{CMISS_EXAMPLES};
#Script/template path
my $template_path = "$example_location/common/cmiss_type1";
my $example_dir = $example_location . "/" . $example_path;
my $example_name = substr ($example_dir, rindex ($example_dir, '/') + 1);
my $run_cvs = "cvs";

if( exists $ENV{CMISS_LOCALE} and $ENV{CMISS_LOCALE} eq "OXFORD" ) {
    my $cvsservername = "esu1.auckland.ac.nz";
    my $cvsroot = "/product/cmiss/cvs";
    my %emailhash = ( 
		      "cpb"=>"bradley", 
		      "mn"=>"nash",
		      "nps"=>"nsmith",
		      "gry"=>"garny",
		      "dpn"=>"nickerso",
		      "pjh"=>"hunter",
		      "ajp2"=>"pullan"
		      );
    my $who = (getpwuid($<))[0];
    if( exists $emailhash{$who} ) {
	$run_cvs .= " -d :pserver:$emailhash{$who}"."@"."$cvsservername:$cvsroot";
    } else {
	die "Username $who is unknown in the email hash\n";
    }
}

my $starting_dir = $ENV{PWD};

if (-d $example_dir)
  {
    die "Directory $example_dir already exists\n";
  }

#Ensure directories exist
my @new_dirs = mkpath("$example_dir", 0, 0755);

#Non testing stuff

mkdir ("$example_dir/cmiss_input", 0755);

open (FILE, ">$example_dir/cmiss_input/example_$example_name.com")
  || die "can't open : $example_dir/cmiss_input/example_$example_name.com: $!\n";
print FILE "# Example $example_name\n";
close FILE;

mkdir ("$example_dir/web_data", 0755);

open (FILE, ">$example_dir/web_data/name.txt")
  || die "can't open : $example_dir/web_data/name.txt: $!\n";
print FILE "Newly created example\n";
close FILE;
open (FILE, ">$example_dir/web_data/description.html")
  || die "can't open : $example_dir/web_data/description.html: $!\n";
print FILE "This is a blank example that hasn't been filled in yet\n";
close FILE;

mkdir("$example_dir/control", 0755)
  || die "Couldn't make $example_dir/control directory\n";
open (FILE, ">$example_dir/control/html")
  || die "can't open : $example_dir/control/html: $!\n";
print FILE <<\EOF;
#!/bin/sh
exec "$CMISS_EXAMPLES/common/cmiss_type1/html.pl" ${1+"$@"}
EOF
close FILE;
chmod (0755, "$example_dir/control/html");

#Cvs
for (@new_dirs)
  {
	 $_ =~ m%([^/]*/?)$% || die "Unable to extract directory and path name.";
	 if (defined $`)
		{
		  chdir($`);
		}
    !system ("$run_cvs add $1") || die "Unable to add directory $1 to cvs.";
  }

chdir($example_dir);
!system ("$run_cvs add cmiss_input/") || die "Unable to add directory cmiss_input/ to cvs.";
!system ("$run_cvs add cmiss_input/example_$example_name.com") || die "Unable to add file $example_dir/cmiss_input/example_$example_name.com to cvs.";
!system ("$run_cvs add web_data/") || die "Unable to add directoryweb_data/ to cvs.";
!system ("$run_cvs add web_data/name.txt") || die "Unable to add file web_data/name.txt to cvs.";
!system ("$run_cvs add web_data/description.html") || die "Unable to add file web_data/description.html to cvs.";
!system ("$run_cvs add control/") || die "Unable to add directory control/ to cvs.";
!system ("$run_cvs add control/html") || die "Unable to add file control/html to cvs.";
if ($cm_testing)
  {
    #Testing
    open (FILE, ">control/test")
      || die "can't open : control/test: $!\n";
    print FILE <<\EOF;
#!/bin/sh
exec "$CMISS_EXAMPLES/common/cmiss_type1/test.pl" ${1+"$@"}
EOF
    close FILE;
    chmod (0755, "control/test");

    open (FILE, ">control/test_run")
      || die "can't open : control/test_run: $!\n";
    print FILE <<\EOF;
#!/bin/sh
exec "$CMISS_EXAMPLES/common/cmiss_type1/test_run_cm.pl" ${1+"$@"}
EOF
    close FILE;
    chmod (0755, "control/test_run");

    open (FILE, ">control/versions")
      || die "can't open : control/versions: $!\n";
    print FILE "!Link to all the cm tests.\n";
    print FILE "#include \"\$(CMISS_EXAMPLES)/common/cmiss_type1/allcm.versions\"\n";
    close FILE;

    open (FILE, ">control/notify_list")
      || die "can't open : control/notify_list: $!\n";
    print FILE "!Link to all the cm programmers.\n";
    print FILE "#include \"\$(CMISS_EXAMPLES)/common/cmiss_type1/allcm.notify_list\"\n";
    close FILE;

    mkdir("testing", 0755)
      || die "Couldn't make testing directory\n";
    mkdir("testing/ndiff", 0755)
      || die "Couldn't make testing/ndiff directory\n";
    mkdir("testing/ndiff/generic", 0755)
      || die "Couldn't make testing/ndiff/generic directory\n";
    mkdir("testing/retain", 0755)
      || die "Couldn't make testing/retain directory\n";
    mkdir("testing/retain/generic", 0755)
      || die "Couldn't make testing/retain/generic directory\n";
    #Make a dummy file so that the log gets copied.
    open (FILE, ">testing/retain/generic/cmiss_test.log")
      || die "can't open : testing/retain/generic/cmiss_test.log$!\n";
    print FILE "\n";
    close FILE;
    
    #Cvs
    !system ("$run_cvs add control/test") || die "Unable to add file control/test to cvs.";
    !system ("$run_cvs add control/test_run") || die "Unable to add file control/test_run to cvs.";
    !system ("$run_cvs add control/versions") || die "Unable to add file control/versions to cvs.";
    !system ("$run_cvs add control/notify_list") || die "Unable to add file control/notify_list to cvs.";
    !system ("$run_cvs add testing/") || die "Unable to add directory testing/ to cvs.";
    !system ("$run_cvs add testing/ndiff/") || die "Unable to add directory testing/ndiff/ to cvs.";
    !system ("$run_cvs add testing/ndiff/generic/") || die "Unable to add directory testing/ndiff/generic/ to cvs.";
    !system ("$run_cvs add testing/retain/") || die "Unable to add directory testing/retain/ to cvs.";
    !system ("$run_cvs add testing/retain/generic/") || die "Unable to add directory testing/retain/generic/ to cvs.";
    !system ("$run_cvs add testing/retain/generic/cmiss_test.log") || die "Unable to add file testing/retain/generic/cmiss_test.log to cvs.";
  }

chdir($starting_dir);
