eval 'exec perl -S $0 ${1+"$@"}'
    if 0;

# Merge `log' history files recieved by send_output.pl from host-local
# directories to the collective output directories, removing the host-local
# copies.

# We could leave files around to save work for rsync next time
# send_output sends files, but there would be no automatic
# mechanism to remove deleted examples.  As this directory is
# hidden people may not notice files lying around


my $usage = <<"END_OF_USAGE;";
Usage: $0 [-h] [host_regexp]
END_OF_USAGE;

use strict;
use Cwd;

# remove path from warning and error messages
$0 =~ s%^.*/%%;
BEGIN {
    $^W = 1;
# # there shouldn't be any warnings
#    $SIG{__WARN__} = sub { die @_ };
    $SIG{__WARN__} = sub { warn "$0 warning: ",@_ };
}
sub dirname ($) {
    # return the directory part of a filename
    $_[0] =~ m'^(.*)/[^/]*$' ? $1 : '.';
}
# use modules in the same directory as this file if they exist.
use lib dirname __FILE__, "$ENV{CMISS_EXAMPLES}/common/cmiss_type1";
use test_cmiss;

use Getopt::Long;

Getopt::Long::GetOptions(
  'h' => sub { print $usage; exit },
 ) or die $usage;

my $examples_root = $ENV{CMISS_EXAMPLES} or
  die "Need to set environment variable CMISS_EXAMPLES\n";

my $host_regexp = @ARGV ? "(?:$ARGV[0])" : '.*';

$SIG{__DIE__} = sub { die "$0: ",@_ };

# Set up a protect mechanism that provides a finish that always
# executes and a catch for dies.

sub protect ( &% ) {
    my ($try,%ops) = @_;
    eval {
 	local $SIG{__DIE__}; # default die
	$try->();
    };
    exists $ops{finish} and do {
	# protect the finish also while retaining @_;
	my $save_at = $@;
	eval { $ops{finish}->() };
	$@ = $save_at . $@;
    };
    $@ and exists $ops{catch} ? $ops{catch}->() : die $@;
}
sub finish (&) { finish => $_[0] };
sub catch (&) { catch => $_[0] };

# my $readdir_grep_string = q/^(\.\.?)$/;

#     # File::Path::rmtree does not provide good error notification.
# #     require File::Path;
#     sub rmtree ($)
#       {
# 	my ($path) = @_;

# 	test_cmiss::recurse_directory_tree
# 	  sub { 1 },
# 	    sub { my ($path,$results,$files) = @_;

# 		  my $result = ! grep { ! $_ } @$results;

# 		  foreach my $file (@$files)
# 		    {
# 		      my $filename = "$path/$file";
# 		      unlink $filename or do
# 			{
# 			  warn "can't unlink $filename: $!\n";
# 			  $result = undef;
# 			};
# 		    }

# 		  rmdir $path or do
# 		    {
# 		      warn "can't rmdir $path: $!\n";
# 		      $result = undef;
# 		    };

# 		  return $result;
# 		},
# 	      $path;
#       }

sub merge_test_logs ($$)
  {
    my ($temp_dir, $keep_dir) = @_;
    # Merges the test history in $temp_dir with that in $keep_dir and removes
    # the one in $temp_dir;
    my $temp_log = "$temp_dir/$test_cmiss::test_log";
    my $keep_log = "$keep_dir/$test_cmiss::test_log";

    return if ! -e $temp_log;

    if( ! -f $keep_log )
      {
	rename $temp_log, $keep_log
	  or die "can't rename $temp_log, $keep_log: $!\n";
	return;
      }

    open KEEP, "+< $keep_log" or die "can't open +< $keep_log: $!\n";
    open TEMP, "< $temp_log" or die "can't open < $temp_log: $!\n";

    # skip header lines
    scalar <KEEP>;
    scalar <TEMP>;

    # Search for the first place where lines of $temp_log are not in
    # $keep_log.

    my ($temp_line,$keep_line,$write_pos);
    my $cmp = 0;		# take a new line from each file

    do { $write_pos = tell KEEP;
	 $keep_line = <KEEP>;
	 if( $cmp == 0 )	# lines are the same
	   { $temp_line = <TEMP> };
       }
      while( defined $keep_line and defined $temp_line
	       and		# keep_line is older or same as temp_line
		 ($cmp = $keep_line cmp $temp_line) <= 0 );

    if( defined $temp_line )
      { # There are new lines that need to be merged.
	my @newlines; # lines that need to be rewritten

	while( defined $keep_line or defined $temp_line )
	  {
	    my $cmp = do
	      { if( ! defined $keep_line )
		  { 1 }		# take temp line
		elsif( ! defined $temp_line )
		  { -1 }	# take keep_line
		else
		  { $keep_line cmp $temp_line }
	      };

	    if( $cmp <= 0 )
	      {	   # keep_line is older (or no temp_line) or same as temp_line
		push @newlines, $keep_line;
		$keep_line = <KEEP>;
		if( $cmp == 0 )	# lines are the same
		  { $temp_line = <TEMP> };
	      }
	    else
	      {			# temp_line older (or no keep_line)
		push @newlines, $temp_line;
		$temp_line = <TEMP>;
	      }
	  }

	seek KEEP, $write_pos, 0 or die "seek failed on $keep_log\n";

	print KEEP @newlines;
      }

    close KEEP or die "close failed on $keep_log: $!\n";
    close TEMP;

    unlink $temp_log
      or warn "can't unlink $temp_log: $!\n";
  }

opendir DIR, $examples_root or die "can't read $examples_root: $!\n";

my @host_dirs = grep { /^.histories-$host_regexp$/ } readdir DIR;

closedir DIR;

foreach my $host_dir (@host_dirs)
  {
    test_cmiss::recurse_directory_tree(
      sub { 1 },
      sub
	{
	  my ($path,$results,$files) = @_;

#	  my $result = ! grep { ! $_ } @$results;

	  foreach my $file (@$files)
	    {
	      my $filename = "$path/$file";

	      $file eq $test_cmiss::test_log
		or die "unexpected file $filename";

	      my $destination = $path;
	      $destination =~
		s|\Q$examples_root/.histories-\E[^/]+/|$examples_root/|
		  or die;

	      protect
		{
		  local $SIG{INT} = sub
		    {
		      die "interrupt received\n";
		    };
		  local $SIG{TERM} = sub
		    {
		      die "SIGTERM received\n";
		    };

		  test_cmiss::start_lock_output $destination;

		  merge_test_logs $path, $destination;

		  test_cmiss::end_lock_output $destination;
		}
		  catch
		    {
		      warn $@;
		      print STDERR "cleaning up locks...";
		      &test_cmiss::abort_locks();
		      die "done\n";
		    };
	    }

	  rmdir $path or do
	    {
	      warn "can't rmdir $path: $!\n";
# 	      $result = undef;
	    };

# 		  return $result;
	},
      "$examples_root/$host_dir"
    );
  }
