#!/usr/bin/perl -w

use strict;
use File::Basename;

use Getopt::Long;

use lib "$ENV{CMISS_EXAMPLES}/common/cmiss_type1";
use test_cmiss;

my $usage = "Usage: $0 [-h] [--verbose] -e example -x cmgui_executable -l stdouterr_filename -s operating_system [--irix-clearmalloc] [--irix-floating-point-trap]\n";

my $example_path = "";
my $executable = 0;
my $log_file = 0;
my %options;
my $operating_system;
my $verbose = 0;
my $valgrind = 0;

if (@ARGV < 1)
  {
    print $usage;
    exit;
  }
Getopt::Long::config qw(bundling no_getopt_compat pass_through no_auto_abbrev);
unless( Getopt::Long::GetOptions
	('h' => sub { print $usage; exit },
	 #Verbose
	 'verbose' => sub { $verbose = 1 },
	 #The relative path to this example i.e. a/a1
	 'e=s' => \$example_path,
	 #Log file.
	 'l=s' => \$log_file,
	 #Executable i.e. /product/cmiss/bin/cm/cm_n32 or my_cmo_n32
	 'x=s' => \$executable,
	 #Operating system
	 's=s' => \$operating_system,
	 #Clear malloced memory
	 'irix-clearmalloc' => sub {
	     $ENV{_SSMALLOC_FASTCHK} = '';
	     $ENV{_SSMALLOC_CLEAR_MALLOC} = '';
	     $ENV{_SSMALLOC_CLEAR_MALLOC_PATTERN} = '0xfffa5a5a';
     },
	 'irix-floating-point-trap' => sub {
		  #Trap Floating Point Excetions.   #Works only when linked with fpe
	     $ENV{TRAP_FPE} = 'DIVZERO=ABORT;INVALID=ABORT';
	   },
	 #valgrind memorycheck testing
	 'valgrind-memorycheck' => \$valgrind
	) ) {
    print STDERR $usage;
    exit 1;
}

if (! $executable)
  {
    print "Executable option missing\n";
    print $usage;
    exit;
  }
if (! $example_path)
  {
    print "Example path option missing\n";
    print $usage;
    exit;
  }
if (! $log_file)
  {
    print "Log file option missing\n";
    print $usage;
    exit;
  }

my $example_location = $ENV{CMISS_EXAMPLES};
my $template_path ="$example_location/common/cmiss_type1";
my $example_name=basename($example_path);
my $example_comfile;
# Try to use resolve_example_path to find the command file to use
my $resolve = "$example_location/common/resolve_example_path";
my @resolved = split ' ',`$resolve $example_path`;
if (@resolved == 2) {
    $example_comfile = $resolved[1];
} else {
    # If resolve_example_path didn't work, just assume things will work.
    if (-f "$example_location/$example_path/cmiss_input/example_$example_name.com") {
        $example_comfile = "example_$example_name";
    } else {
        $example_comfile = "$example_name";
    }
}
my $temp_comfile_name = "cmgui_test.com";

open(TEMP_COM,">$temp_comfile_name") || die "Error opening file $temp_comfile_name, $!\n";

# Need to quote the example_path to avoid paths like 1/11/113
# being treated as a series of divisions.
print TEMP_COM "set directory example \"$example_path\"\n";

# Set the $TESTING flag to 1 in the command file. Comfiles may put condition
# "if ($TESTING) ..." around commands they only want to run during testing.
print TEMP_COM "\$TESTING = 1;\n";
if (defined $operating_system)
  {
	 print TEMP_COM "\$OPERATING_SYSTEM = \"$operating_system\";\n";
  }

open(PART,"<$example_location/$example_path/cmiss_input/$example_comfile.com") || die "Error opening file $example_location/$example_path/cmiss_input/$example_comfile.com, $!\n";
print TEMP_COM $_ while( <PART>);
close(PART) || die;

close(TEMP_COM) || die;

my $command="$executable -batch -no_display -epath $example_location $temp_comfile_name";
if ($valgrind)
  {
	 my $valgrind_executable;
	 my $valgrind_version;

	 if (-f "$ENV{CMISS_ROOT}/valgrind/bin/valgrind")
		{
		  $valgrind_executable = "$ENV{CMISS_ROOT}/valgrind/bin/valgrind";
		}
	 else
		{
		  $valgrind_executable = "valgrind";
		}
	 $valgrind_version = `$valgrind_executable --version`;
	 if ($valgrind_version && ($valgrind_version =~ m/valgrind-(\d)\.(\d)\.\d/))
	 {
		my $valgrind_major_version = $1;
		my $valgrind_minor_version = $2;
		if (($valgrind_major_version > 2) ||
			 (($valgrind_major_version == 2) && ($valgrind_minor_version > 3)))
		{
		  $command = "$valgrind_executable --tool=memcheck --leak-check=yes --show-reachable=yes --quiet $command";
		}
		else
		{
		  die "At least valgrind 2.4.0 is required.";
		}
	 }
	 else
	 {
		die "memorycheck valgrind version testing requires valgrind to be installed.";
	 }
  }

if ($operating_system eq "irix")
  {
	 $command="(export LD_LIBRARYN32_PATH=\"$ENV{CMISS_ROOT}/mesa/lib/mips-n32-irix/\"; $command)";
  }
elsif ($operating_system eq "irix64")
  {
	 $command="(export LD_LIBRARY64_PATH=\"$ENV{CMISS_ROOT}/mesa/lib/mips-64-irix/\"; $command)";
  }
elsif ($operating_system eq "linux")
  {
	 $command="(export LD_LIBRARY_PATH=\"$ENV{CMISS_ROOT}/mesa/lib/i686-linux/\"; $command)";
  }
elsif ($operating_system eq "x86_64-linux")
  {
    #Work around a locale problem that has appeared on bioeng22 running Ubuntu where
    #Motif generates some errors with fonts if using the en_NZ LANG and LANGUAGE.
    $ENV{LC_ALL} = "C";
	 $command="(export LD_LIBRARY_PATH=\"$ENV{CMISS_ROOT}/mesa/lib/x86_64-linux/\"; $command)";
  }
elsif ($operating_system eq "aix")
  {
	 $command="(export LIBPATH=\"$ENV{CMISS_ROOT}/mesa/lib/rs6000-32-aix/\"; $command)";
  }
elsif ($operating_system eq "aix64")
  {
	 $command="(export LIBPATH=\"$ENV{CMISS_ROOT}/mesa/lib/rs6000-64-aix/\"; $command)";
  }
if ($verbose)
  {
	 print "Executing $command\n";
  }
my $return_code = system("$command < /dev/null > $log_file 2>&1");

#Propogate the return code or signal of the command unless it is 0x100 which
#is success from cmgui
if ($return_code == 0x100)
  {
	 $return_code = 0;
  }
if ($return_code & 0xff)
  {
	 kill ($return_code & 0xff, $$) or die "Kill failed $$ $!";
  }
else
  {
	 exit ($return_code >> 8);
  }











