eval 'exec perl -wS $0 ${1+"$@"}'
    if 0;

# Alternative header that does not work well with ddd
#  #!/bin/sh
#  # The following exec statement causes the first perl in $PATH to
#  # interpret the Perl script beginning at #!perl.  If you want to use
#  # another perl interpreter you'll need to provide the -x option to the
#  # interpreter.  Otherwise a perl interpreter will still exec /bin/sh
#  # on this file and the same behaviour will occur.
#  # To use ddd on the Perl I think `ddd --perl -x' is easiest,
#  # but it gets confused about which line you are at.
#  exec perl -x -S $0 ${1+"$@"}
#  #!perl -w

use strict;
#use diagnostics;
use FileHandle;

sub dirname ($) {
    # return the directory part of a filename
    $_[0] =~ m%^(.*)/[^/]*$% ? $1 : '.';
}
# use modules in the same directory as this file if they exist.
use lib dirname __FILE__, "$ENV{CMISS_EXAMPLES}/common/cmiss_type1";
use test_cmiss;

use Getopt::Long;
use Text::ParseWords;
use File::Path;

my $usage = <<EOF;
Usage: $0 [-h | --help]
       [-s operating_system_regexp]
       [-e example_subpath_list]
       [-p program_name_regexp]
       [-v version_name_regexp
         [-o output_version | --output-version=output_version]]
       [-x executable
         [-c compare_version]]
       [--check-all-broken-versions]
       [--output-root=root_of_output_tree]
       [-k | --keep-all-output]
       [-z number_of_concurrent_processes]
EOF
my $operating_system_regexp;
my %operating_system_regexps = ('IRIX64' => 'irix(?:64)?');
my @subpaths = ();
my $program_name_regexp = ".*";
my $version_name_regexp;
my $explicit_executable;
my $explicit_compare_version;
my $keep_all_output = 0;
my $check_broken_versions = 0;
my $verbose = 0;
# resources available
my %resources_free = ( cpu => 1, memory => 256 );
my $output_root;
my $explicit_output_version;
my $test_summary = $test_cmiss::test_summary;

#Process the command line
Getopt::Long::config qw(bundling no_getopt_compat no_auto_abbrev);
unless( Getopt::Long::GetOptions
	('h|help' => sub { print $usage; exit },
	 #Verbose
	 'verbose' => sub { $verbose = 1 },
	 #Operating system pattern i.e. irix, linux or irix64
	 's=s' => \$operating_system_regexp,
	 #Colon or space separated relative paths to example trees
	 # e.g. a/a1:b/b2
	 'e=s' => sub { push @subpaths, split /[: ]/, $_[1] },
	 #Program name pattern i.e. cmgui or cm
	 'p=s' => \$program_name_regexp,
	 #Version name pattern
	 # e.g. cmgui, cmgui_optimised, cm_n32, cm_64 or cmo_n32
	 'v=s' => \$version_name_regexp,
	 #Executable i.e. /product/cmiss/bin/cmgui or my_cmo_n32
	 #If executable is specified then this overrides that specified
	 # in the versions file.
	 'x=s' => \$explicit_executable,
	 #Compare_version i.e. generic, cmgui_optimised or hpc
	 #If compare_version is specified then this overrides that specified
	 # in the versions file.
	 'c=s' => \$explicit_compare_version,
	 #Number of concurrent executions.
	 'z=s' => \$resources_free{cpu},
	 #Check all broken versions of these examples
	 'check-all-broken-versions' => sub { $check_broken_versions = 1 },
	 #Keep all the output from the test runs in the output directories
	 'k|keep-all-output' => sub { $keep_all_output = 1 },
	 #The root of the output tree
	 'output-root=s' => \$output_root,
	 #The subdirectory into which the output will be loaded.
	 'o|output-version=s' => \$explicit_output_version,
	 #Resource limits (Mutexes (MUTual EXclusion)).
	 'M=s' => sub {
	     my $opts=$_[1];
	     while( $opts =~ m/ *(\w+)=(\d+),?/g ) {
		 $resources_free{$1} = $2;
	     }
	 } ) ) {
    print STDERR $usage;
    exit 1;
}
# Setup for GetOptions call in parse_job_requirements
Getopt::Long::config qw(pass_through);

if (@ARGV)
  {
    print STDERR $usage;
    exit 1;
  }

if ($explicit_compare_version && !($explicit_executable))
  {
    print STDERR "Must specify an executable (-x) if you want to override compare version (-c)\n";
    print STDERR $usage;
    exit 1;
  }

sub abort {
    die "$0: ",@_;
}
sub mywarn {
    warn "$0 warning: ",@_;
}

$SIG{__DIE__} = \&abort;
$SIG{__WARN__} = \&mywarn;

unless (defined $operating_system_regexp) {
    $operating_system_regexp = &operating_system_regexp();
}

unless( @subpaths ) { @subpaths = ('') }

#Where all the examples are stored on disk i.e. /product/cmiss/examples
my $example_location = $ENV{CMISS_EXAMPLES};
defined $example_location or die "set CMISS_EXAMPLES environment variable\n";
my $readdir_grep_string = q/^(\.|CVS)/;

#Check that the current directory is empty.
opendir(DIR, ".") || die "can't opendir current working directory: $!\n";
my @current_dir = grep { ! /$readdir_grep_string/ } readdir(DIR);
closedir DIR;
# if (@current_dir)
# {
# #  die "Current working directory is not empty";
#   print "Warning: current working directory is not empty\n";
# }

if( defined $explicit_output_version and $explicit_output_version =~ m%^/% )
  {
    die <<EOF;
output_version should be a relative directory or path.
Use --output_root to specify an absolute path.
$usage
EOF
  }

if( defined $output_root )
  {
    # Convert to a full path for safety
    $output_root =~ m%^/% or
      $output_root = do
	{
	  require Cwd;
	  Cwd::getcwd() . "/$output_root";
	};
  }
else
  { # $output_root not defined
    $output_root = do
      {
	if ( $explicit_executable && ! $explicit_output_version)
	  {
	    # The normal executable is not used so make it difficult to put the
	    # output in the normal place.  (It can still be done using
	    # --output-root or --output-version explicitly.)

	    # Make an output directory right here
	    require Cwd;
# 	if (! -d $local_output_dir)
# 	  {
# 	    mkdir($local_output_dir, 0755)
# 	      || die "Can't make $local_output_dir directory: $!\n";
# 	  }
	    Cwd::getcwd() . "/cmiss_output";
	  }
	else
	  {
	    $example_location;
	  }
      };
  }

if( ! defined $version_name_regexp )
  {
    defined $explicit_output_version and
      die "Specify a version to test (-v) for this output_version.\n$usage";

    $version_name_regexp = do
      {
	if ($check_broken_versions)
	  { # Do not match any versions automatically, only check broken ones.
	    "";
	  }
	else
	  { # Test all versions
	    ".*";
	    }
      };
  }

my $finish_time = exists $resources_free{time} &&
  ( $^T + $resources_free{time} );

sub resources_are_available ($)
  {
    my ($job_resources) = @_;

    foreach my $resource (keys %$job_resources){
      if( defined $resources_free{$resource}
	    and $job_resources->{$resource} > $resources_free{$resource} )
	{
	  return undef;
	}
    }
    return 1;
  }


sub mark_test_incomplete( $ )
  {
    # Make the Summary file say that to job is scheduled to
    # run.  This will get overwritten when the job is
    # completed.
    my ($job) = @_;
    my ($full_output_directory,$compare_version) =
      @$job{'full_output_directory','compare_version'};

    unless (-d "$full_output_directory")
      {
	mkpath($full_output_directory, 0, 0755)
	  || die "Couldn't make $full_output_directory directory: $!\n";
      }
    my $summary_file = "$full_output_directory/$test_summary";
    my $login = getlogin || '*unknown*';
    my $date = localtime(time());

    open TEST_FILE, ">$summary_file" or
      die "can't open >$summary_file: $!\n";
    print TEST_FILE "***** CMISS_testing $test_summary test_tree.pl $compare_version Pending *****\n";
    print TEST_FILE "Clock_execution running seconds.  CPU times running running\n";
    print TEST_FILE "Test job started by $login $date\n";
    close TEST_FILE or die "error closing $summary_file: $!\n";
  }


sub execute_job( $ ) {
    #This function executes a single job.
    my ($job) = @_;

    if( $finish_time )
      {
	mark_test_incomplete $job
      }
    # else already marked incomplete

    #Make a directory for this example to test in. cf. wait_for_job

    my $cmiss_test_dir = "cmisstest${$}j$job->{id}";

    # If we have to come up with another name then we need to record
    # this for cleanup when the process finishes.
#     while (-d $cmiss_test_dir)
#       {
# 	$cmiss_test_dir = "${cmiss_test_dir}x";
#       }

    mkdir $cmiss_test_dir, 0755
	or die "can't make $cmiss_test_dir directory: $!\n";

    my $command = $job->{command};
	 if ($verbose)
		{
		  print "@$command\n";
		}

    defined( my $pid = fork ) or die "can't fork: $!\n";
    $pid and return $pid;

    # child

    my $script = $command->[0];

    # Alter the umask for generated files according to read
    # permissions on the script.  We should really remove group
    # read privilege if other read is not set and we are in a
    # different group but CVS doesn't bother with this so
    # there doesn't seem much point.
    stat $script or die "can't stat $script: $!\n";
    my $mode = ( stat _ )[2];
    umask( umask | ( ~$mode & 0044 ) );

    #Use own working directory
    chdir $cmiss_test_dir or die "can't chdir to $cmiss_test_dir: $!\n";
    
    exec @$command or die "can't exec $script: $!\n";
}

# key-indexed list of jobs
# hash{id}->{qw(id subpath version command required resources)}
my %jobs_waiting;
# priority reverse-ordered list of jobs
my @jobs_waiting;

my %pids_running;
my $job_id = 1;

sub add_job_to_hash($$) {
    my ($job,$pid) = @_;

    # add stuff to mutexes
    my $job_resources = $job->{resources};
    foreach my $resource (keys %$job_resources){
	if( defined $resources_free{$resource} )
	  {
	    # The time resource is special in that running a job does
	    # not mean the resource is consumed as other jobs may be run
	    # in parallel.
	    if( $resource ne 'time' )
	      {
		$resources_free{$resource} -= $job_resources->{$resource};
	      }
	  }
    }
    $pids_running{$pid} = $job;
}

sub start_jobs()
{
  TRYJOB: for( my $job_index = @jobs_waiting; $job_index-- > 0;  )
    {
	my $tryjob = $jobs_waiting[$job_index];
	### Check if we need to wait on other jobs
	### either in the wait queue or running
	foreach my $wait_for_job (@{$tryjob->{required}}) {
	    exists $jobs_waiting{$wait_for_job} and next TRYJOB;
	    foreach my $running_job (values %pids_running) {
		$wait_for_job == $running_job->{id} and next TRYJOB;
	    }
	}
	### check have enough procs and resources
	resources_are_available $tryjob->{resources} or next TRYJOB;

	print "Starting $tryjob->{subpath}:$tryjob->{version}\n";
	my $pid = execute_job $tryjob;
	add_job_to_hash $tryjob, $pid;
	### remove job from waiting lists
	splice @jobs_waiting, $job_index, 1;
	delete $jobs_waiting{$tryjob->{id}};

	last unless $resources_free{cpu} > 0;
    }

    if( %pids_running )
      {
	return 1;
      }
    else
      {
	print "Not enough resources available for ",
	  (join " ", map { "$_->{subpath}:$_->{version}" } @jobs_waiting),
	    "\n";
	return undef;
      }
}

sub wait_for_job() {
    my $pid = wait;
    if( $pid == -1 ) {
	die "child(ren) have disappeared\n";
    }
    exists $pids_running{$pid} or do {
	warn "don't know about process $pid\n";
	return;
    };

    my $job = $pids_running{$pid};
    my $example = "$job->{subpath}:$job->{version}";

    # Historic behaviour has been to leave the directory if the test
    # script fails.  Perhaps this may leave evidence of the problem.
    if( $? & 0xFF ) {
	warn "$job->{command}[0] received signal: $?\n";
    } elsif(( $? != 0 ) && ( $? != (1 << 8) )) {
	warn "$job->{command}[0] failed: ", $? >> 8, "\n";
    } else {
	#Empty out and remove the test directory.
	my $cmiss_test_dir = "cmisstest${$}j$job->{id}";
	if( ! opendir DIR, $cmiss_test_dir) { 
	    warn "can't opendir $cmiss_test_dir directory: $!";
	} else {
	    @current_dir = grep { ! /$readdir_grep_string/ } readdir(DIR);
	    closedir DIR;
	    for (@current_dir)
	    {
		unlink "$cmiss_test_dir/$_";
	    }
	    rmdir $cmiss_test_dir or warn "can't rmdir $cmiss_test_dir: $!\n";
	}
	if ($? == 0)
	  {
		 print "Finished example $example : SUCCESS\n";
	  }
	else
	  {
		 print "Finished example $example : FAILURE\n";
	  }
    }

    # deallocate/update resources;
    my $job_resources = $job->{resources};
    foreach my $resource ( keys %$job_resources ) {
	if( defined $resources_free{$resource} )
	  {
	    if( $resource ne 'time' )
	      {
		$resources_free{$resource} += $job_resources->{$resource};
	      }
	  }
    }
    if( $finish_time )
      {
	$resources_free{time} = $finish_time - time;
      }
    delete $pids_running{$pid};

    if( %pids_running ) {
	print "examples running: ";
	foreach my $job (values %pids_running){
	    print "$job->{subpath}:$job->{version} ";
	}
	print "\n";
    }
}

sub parse_job_requirements ($@)
  {
    my ($job,@flags) = @_;
    # Parses flags for a job and includes the requirements in the job hash.

    local @ARGV = @flags;

    my @required_jobs = ();
    my %resources = ( cpu => 1 );

    Getopt::Long::GetOptions
	(#Number of processes for this job
	  'p=i' => \$resources{cpu},
	  #Resource usage for this job.
	  'M=s' => sub
	    {
	      my $opts=$_[1];
	      while( $opts =~ m/ *(\w+)=(\d+),*/g ) {
		$resources{$1} = $2;
	      }
	    },
	  #This flag specifies that this version can only be started once
	  #the -w list of other versions is complete.  This is intended to be
	  #used to compare different versions, say multiprocessing speedups.
	  'w=s' => sub
	    {
	      my @required_versions = split (",", $_[1]);
	      # !!! This algorthm is very slow and could be improved.
	      # How many examples have a -w flag?
	      foreach my $required_version (@required_versions) {
		foreach my $otherjob (@jobs_waiting) {
		  if( $otherjob->{version} eq $required_version
			and $otherjob->{subpath} eq $job->{subpath} )
		    {
		      push @required_jobs, $job->{id};
		      last;
		    }
		} continue {
		  die "Unable to find required version for $job->{command}\n";
		}
	      }
	    }
	 ) or warn "don't understand options for $job->{command}\n";

    $job->{resources} = \%resources;
    $job->{required} = \@required_jobs;
  }

sub load_job ($) {
    my $job = shift;
    #This function adds one job to the set of jobs to run.

    $job->{id} = $job_id;
    push @jobs_waiting, $job;
    $jobs_waiting{$job_id} = $job;
    $job_id++;
}

sub read_versions( $ ); # Extra Prototype due to recursion

sub read_versions( $ ) {
    my ($versions_filename) = @_;

    # Read all the versions from a versions file in such a way that
    # there is only one test for each version.  If a version is found
    # more than once then the last supercedes the former.  This provides a
    # mechanism to include default files then override some versions.

    # Returns a reference to a hash keyed by version name of arrays
    # containing the words in each execution line of the versions file
    # or included files.

    my %versions;

    my $versions_file = new FileHandle("<$versions_filename") or do
      {
	warn "can't open $versions_filename: $!\n";
	return;
      };

    while (<$versions_file>)
      {
	if (m"^[!\n]")
	  {
	    #Comment or blank line, do nothing
	  }
	elsif (m"^#")
	  {
	    #Directive
	    if (m%^#include "(.*)" *$%)
	      {
		my $filename = $1;
		$filename =~ s%\$[(](.*?)[)]%$ENV{$1}%g;
		if( my $new_versions = read_versions $filename )
		  {
		    while( my ($version_name,$fields) = each %$new_versions )
		      {
			$versions{$version_name} = $fields;
		      }
		  }
	      }
	    else
	      {
		warn "Unknown directive \"$_\" in file $versions_filename";
		next;
	      }
	  }
	else
	{
	    #Assume this is an execution line
	    # line has fields, OS, Program Name, Version Name, Executable,
	    # Compare Version, and optional Flags
	    my @fields = split;

	    # Check that the line of the versions file has enough fields
	    unless( @fields >= 5 ) {
		warn "don't understand line $. of $versions_filename\n";
		next;
	    }

	    my ($operating_system,$program_name,$version_name) = @fields[0..2];

	    # Only override the a previous definition of the version if the
	    # program name matches those that are going to be tested.

	    # Brackets around regexps ensure that ^ and $ function
	    # around both parts of | operators.
	    if( $operating_system =~ m/^(?:$operating_system_regexp)$/i
		  && $program_name =~ m/^(?:$program_name_regexp)$/ )
	      {
		$versions{$version_name} = \@fields;
	      }
	}
      }
    close ($versions_file);

    return \%versions;
  }


sub load_versions( $$$ )
  {
    my ($versions,$example_dir,$subpath) = @_;

	 my $version_matches;
	 #The first part of this routine determines if this is represents
	 #a version that should be tested.  If the $test_this_version flag
    #is set here then the rest of the routine marks the example as being
    #tested and queues it for processing.
	 my $test_this_version;

    # Set up jobs for each version in the hash reference $versions for
    # the example $subpath that matches those selected by the user.

    # !!! If $explicit_output_version is specified we should probably check
    # that only one version is tested or we will loose the output from the
    # others.

    foreach my $version ( map { $versions->{$_} }
			    sort { $a cmp $b } keys %$versions )
      {
		my ($operating_system,$program_name,$version_name,$executable,
			 $compare_version,@flags) = @$version;

		$test_this_version = 0;
		# operating system and program name have already been checked.
		$version_matches = ($version_name =~ m/^(?:$version_name_regexp)$/);
		if ( $version_matches || $check_broken_versions ) {
		  #If we are checking broken versions the we will look at the summary file
		  if ($version_matches)
			 {
				$test_this_version = 1;
			 }
		  my $output_version =
		    defined $explicit_output_version ? $explicit_output_version
		      : $version_name;

		  if ($explicit_executable) {
		    $executable = $explicit_executable;
		    if ($explicit_compare_version)
		      {
			$compare_version = $explicit_compare_version;
		      }
		  } else {
		    # variable substitution may also be required on the other
		    # components.
		    $executable =~ s/(^|[^\\])\$(\w+)/$1$ENV{$2}/;
		    $executable =~ s/(^|[^\\])\${(\w+)}/$1$ENV{$2}/;
		  }

		  my $full_output_directory =
		    "$output_root/$subpath/output/$output_version";

		  my $break_and_success_hash_reference = 
          test_cmiss::last_break_and_success($full_output_directory);
        my $last_success =
		    ${ $break_and_success_hash_reference }{success};
        my $last_break =
		    ${ $break_and_success_hash_reference }{break};
		  my $last = $break_and_success_hash_reference->{last};

		  # Estimate of time required based on previous run:
		  # (Using 0 if there has been no previous test means that
		  # such examples will run first.)
		  my $time_estimate = $last->{'real time'} || 0;

		if( $last_success )
		  {
		    my $success_time = $last_success->{'real time'};
		    # Record the previous successful execution time
		    push @flags, '-t', $success_time;
		    # Use this value to adjust $time_estimate if it is larger.
		    if( $success_time > $time_estimate )
		      { $time_estimate = $success_time }
		  }

      if ( $check_broken_versions )
      {
         #If there is a break and last run was a break
         if ( $last_break && ((! $last_success) ||
           ($last_success->{'start time'} < $last_break->{'start time'} )))
        {
          $test_this_version = 1;
        }
      }

	    if ($test_this_version)
	      {
		if ($keep_all_output)
		  {
		    push @flags, '-k';
		  }

		if ($verbose)
		  {
			 push @flags, '--verbose';
		  }

		my $job = { subpath => $subpath, version => $version_name,
			    command => ["$example_dir/control/test",
					'-e',$subpath,
					'-s',$operating_system,
					'-x',$executable,
					'-c',$compare_version,
					'-o',$full_output_directory,
					@flags],
			    full_output_directory => $full_output_directory,
			    compare_version => $compare_version
			   };

		parse_job_requirements $job, @flags;

 		$job->{resources}{time} = $time_estimate;

		# Check whether there are actually enough resources to
		# run the job.
		if( ! resources_are_available $job->{resources} )
		  {

		    print
"Not enough resources available for $job->{subpath}:$job->{version}\n";
		  }
		else
		  {
		    # This priority calculation means that each example is
		    # entitled to the same quantity of cpu resources.
		    # lower priorities are run first.
		    $job->{priority} =
		      exists $last->{'real time'} ?
			( $last->{'real time'} || 1 ) # round up if less than 1
			  * $job->{resources}{cpu}
			    / ( ( $^T - $last->{'start time'} ) || 1 )
			      : 0;

		    load_job $job;

		    # If the number of jobs run is constrained by a time limit
		    # then we don't yet know which jobs will actually be run
		    # so don't yet leave anything to say that the test is
		    # incomplete.  (This could be done when the jobs is run.)
		    unless( $finish_time )
		      {
			mark_test_incomplete $job;
		      }
		  }

	    }
	 }
    }
  }

#Recurse down tree.
sub load_tests_tree
  {
    my $subpath = shift;
    my $example_dir = $example_location . "/" . $subpath;
    my @subdirectories;
    
    if (test_cmiss::is_leaf_directory("$example_dir"))
      {
	#Read the versions file and see if we have some matches.
	if ((-f "$example_dir/control/versions") && 
	    (-f "$example_dir/control/test"))
	  {
	    if( my $versions = read_versions "$example_dir/control/versions" )
	      { load_versions $versions, $example_dir, $subpath }
	  }
      }
    else
      {
	#Look further down.
	
	opendir(DIR, "$example_dir") || die "can't opendir $example_dir: $!";
	@subdirectories = sort {$a cmp $b} grep { ! /$readdir_grep_string/ } readdir(DIR);
	closedir DIR;
	
	for (@subdirectories)
	  {
	    if (-d "$example_dir/$_") {
		load_tests_tree
		    ( length($subpath) > 0 ? "$subpath/$_" : $_ );
	    }
	  }
      }
  }

foreach my $subpath ( @subpaths ) {
    load_tests_tree $subpath;
}

if (! scalar @jobs_waiting)
  {
	 die ("No jobs for OS \"$operating_system_regexp\" and program \"$program_name_regexp\" matching version \"$version_name_regexp\"");
  }

# Sort the jobs so that the shortest run first.  This way we get more results
# sooner.  If no times are available run in the order found.  The list is in
# reverse order of execution because it is then easier to splice the array to
# remove jobs.
@jobs_waiting =
  sort { $b->{priority} <=> $a->{priority} || $b->{id} <=> $a->{id}
       } @jobs_waiting;

while( @jobs_waiting and start_jobs )
  {
    wait_for_job;
  }
while( %pids_running )
  {
    wait_for_job;
  }

if ($output_root ne $example_location)
  {
    my $mail_notify = "ALL";

    print "\n\n";

    #Print mail results to screen for output generated.
    my ($notify_hash_ref,$failure_list_ref) =
	test_cmiss::make_mail_trees( $output_root, $example_location,
				     \@subpaths, undef, [$mail_notify] );


    unless( exists $notify_hash_ref->{$mail_notify})
    {
	print "No test examples in @subpaths!\n";
    }
    elsif (@$failure_list_ref)
    {
	foreach my $failure_ref (@$failure_list_ref) {
	    print
"Failure for example $failure_ref->{'example path'}: $failure_ref->{name}\n";
	    my $failed_versions_ref = $failure_ref->{'failed versions'};
	    my $incomplete_versions_ref = $failure_ref->{'incomplete versions'};
	    if( @$failed_versions_ref ) {
		print "Failed versions: @$failed_versions_ref\n";
	    }
	    if( @$incomplete_versions_ref ) {
		print "Incomplete versions: @$incomplete_versions_ref\n";
	    }
	}
	print "See files in $output_root\n";
      }
    else
      {
	print "All tested @subpaths examples were successful!\n";
      }
  }

sub operating_system_regexp {
    require POSIX;
    my ($sysname,$machname) = (&POSIX::uname())[0,4];
    if( $sysname eq 'Linux' && $machname !~ 'i.86' ) {
      # Linux is available on many platforms so be more specific.
      $sysname = "$machname-$sysname";
    }
    return exists $operating_system_regexps{$sysname} ?
	$operating_system_regexps{$sysname} : $sysname;
}

### Local Variables: 
### tab-width: 8
### End: 
